package meerkat;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.bulletinboard.SimpleBulletinBoardClient;

import meerkat.comm.CommunicationException;
import meerkat.crypto.Encryption;
import meerkat.protobuf.BulletinBoardAPI;
import meerkat.protobuf.BulletinBoardAPI.BulletinBoardMessage;
import meerkat.protobuf.BulletinBoardAPI.MessageFilter;
import meerkat.protobuf.BulletinBoardAPI.MessageFilterList;
import meerkat.protobuf.BulletinBoardAPI.UnsignedBulletinBoardMessage;
import meerkat.protobuf.Crypto;
import util.AccurateTimestamp;
import util.RegistryTags;

import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Random;

import static util.CollectionMessagesUtils.*;

/**
 * Created by Vladimir Eliezer Tokarev on 1/8/2016.
 * Gives the ability to synchronously manage voters information
 */
public class SimpleRegistry  implements RegistryInstance{

    protected Encryption signatory;
    protected SimpleBulletinBoardClient communicator;

    public SimpleRegistry(Encryption signatory, SimpleBulletinBoardClient communicator) {
        this.signatory = signatory;
        this.communicator = communicator;
    }

    /**
     * Creates BulletinBoardMessage with signed basicMessage and UnsignedBulletinBoardMessage that contains the basic message
     *
     * @param basicMessage BasicMessage
     * @return BulletinBoardMessage
     */
    private BulletinBoardMessage CreateBulletinBoardMessage(UnsignedBulletinBoardMessage basicMessage) {
        try {
            BulletinBoardMessage.Builder bulletinBoardMessage =
                    BulletinBoardMessage.newBuilder();

            Crypto.RerandomizableEncryptedMessage encryptedMessage =
                    signatory.encrypt(basicMessage, signatory.generateRandomness(new Random()));

            Crypto.Signature.Builder messageSignature = Crypto.Signature.newBuilder();
            messageSignature.setType(Crypto.SignatureType.ECDSA)
                    .setData(encryptedMessage.toByteString());

            bulletinBoardMessage.setMsg(basicMessage);

            bulletinBoardMessage.addSig(messageSignature);

            return bulletinBoardMessage.build();
        } catch (IOException e) {
            return null;
        }
    }

    /**
     * Gets messages that have the wanted id tag and have or the ADD_TO_GROUP_TAG or REMOVE_FROM_GROUP_TAG
     *
     * @param tags the tags based on which the messages will be filtered
     * @return MessageFilterList.
     */
    public MessageFilterList GetRelevantMessages(List<String> tags) {
        MessageFilterList.Builder filters = MessageFilterList.newBuilder();

        if (tags == null){
            return filters.build();
        }

        for (String tag : tags) {
            MessageFilter.Builder filter =
                    MessageFilter.newBuilder().setTag(tag)
                            .setType(BulletinBoardAPI.FilterType.TAG);

            filters.addFilter(filter);
        }
        return filters.build();
    }

    /**
     * Gets Relevant bulletinBoard messages from communicator than converts them to VoterRegistryMessages
     *
     * @param tags list of tags that will be used as filters
     * @return List<VoterRegistryMessage>
     */
    private List<VoterRegistryMessage> GetRelevantVoterRegistryMessages(List<String> tags) throws InvalidProtocolBufferException {
        List<BulletinBoardMessage> relevantMessages = communicator.readMessages(GetRelevantMessages(tags));
        List<UnsignedBulletinBoardMessage> messages = GetUnsignedBulletinBoardMessages(relevantMessages);
        return ConvertToVoterRegistryMessages(messages);
    }

    /**
     * Tries to post basicMessage and return true if its been successfully posted
     *
     * @param Message the massage to post
     * @return true when the post was successful else false
     */
    private boolean SafePost(BulletinBoardMessage Message) {
        try {
            communicator.postMessage(Message);
            return true;
        } catch (CommunicationException e) {
            return false;
        }
    }

    public void AddVoter(String voterID, String personalData, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage.Builder basicMessage =
                UnsignedBulletinBoardMessage.newBuilder().
                        addTag(RegistryTags.ID_TAG + voterID)
                        .addTag(RegistryTags.VOTER_ENTRY_TAG.toString())
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + AccurateTimestamp.GetCurrentTimestampString());

        basicMessage.setData(ByteString.copyFrom(personalData.getBytes()));

        callback.HandleVoterAdded(SafePost(CreateBulletinBoardMessage(basicMessage.build())));
    }

    public void AddToGroup(String voterID, String groupID, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage.Builder basicMessage =
                UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + voterID)
                        .addTag(RegistryTags.GROUP_ID_TAG + groupID)
                        .addTag(RegistryTags.GROUP_ACTION_TAG .toString() + RegistryTags.ADD_TO_GROUP_TAG)
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + AccurateTimestamp.GetCurrentTimestampString());

        callback.HandleVoterAddedToGroup(SafePost(CreateBulletinBoardMessage(basicMessage.build())));
    }

    public void RemoveFromGroup(String voterID, String groupID, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage.Builder basicMessage =
                UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + voterID)
                        .addTag(RegistryTags.GROUP_ID_TAG + groupID)
                        .addTag(RegistryTags.GROUP_ACTION_TAG.toString()+ RegistryTags.REMOVE_FROM_GROUP_TAG)
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + AccurateTimestamp.GetCurrentTimestampString());

        callback.HandleVoterRemovedFromGroup(SafePost(CreateBulletinBoardMessage(basicMessage.build())));
    }

    public void SetVoted(String id, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage.Builder basicMessage =
                UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + id)
                        .addTag(RegistryTags.VOTE_ACTION_TAG.toString())
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + AccurateTimestamp.GetCurrentTimestampString());

        callback.HandleVoterAddedToGroup(SafePost(CreateBulletinBoardMessage(basicMessage.build())));
    }

    public void GetGroups(String id, RegistryCallBack callback) {
        try {
            List<String> GroupsActionsTags = new ArrayList<String>() {{
               add(RegistryTags.GROUP_ID_TAG + id);
            }};
            List<VoterRegistryMessage> voterRegistryMessages = GetRelevantVoterRegistryMessages(GroupsActionsTags);

            Map<String, VoterRegistryMessage> groupIdToMessage = GetLatestGroupsActions(voterRegistryMessages);

            callback.HandleVoterGroups(GetListOfGroupIds(groupIdToMessage));
        } catch (ParseException | InvalidProtocolBufferException e) {
            callback.HandleVoterGroups(null);
        }
    }

    public void GetPersonIDDetails(String id, RegistryCallBack callback) {
        try {
            List<String> GroupsActionsTags = new ArrayList<String>() {{
                add(RegistryTags.ID_TAG + id);
                add(RegistryTags.VOTER_ENTRY_TAG.toString());
            }};
            List<VoterRegistryMessage> voterRegistryMessages = GetRelevantVoterRegistryMessages(GroupsActionsTags);

            VoterRegistryMessage LatestMessage = GetLatestMessage(voterRegistryMessages);

            callback.HandleVoterInfo(LatestMessage.tagsToStringList());
        } catch (InvalidProtocolBufferException | ParseException e ){
            callback.HandleVoterInfo(null);
        }
    }
}
